function TypedPromises(valueOrPromise) {
    return TypedPromises.promisify(valueOrPromise);
}

/**
*	A module with functions for creating deferred objects that can be resolve and with other useful related functions.
*/
var TypedPromises;
(function (TypedPromises) {
    /**
    *	Checks if a value is a function.
    *	http://stackoverflow.com/questions/5999998/how-can-i-check-if-a-javascript-variable-is-function-type
    *	@param f A value that may or may not be a function.
    */
    function isFunction(f) {
        var getType = {};
        return !!f && getType.toString.call(f) === '[object Function]';
    }
    TypedPromises.isFunction = isFunction;

    /**
    *	Checks if a value is a promise - i.e. has a then function
    *	@param obj A value that may or may not be a promise.
    */
    function isPromise(obj) {
        return obj && obj.then && isFunction(obj.then);
    }
    TypedPromises.isPromise = isPromise;

    /**
    *	Checks if a value is a promise from this library. Used to enforce conformance to APlusPromise standard.
    * @param obj A value that may or may not be an AGPromise.
    */
    function isAGPromise(obj) {
        return obj && obj._isAGPromise;
    }
    TypedPromises.isAGPromise = isAGPromise;

    /**
    *	This function defers a function to run in a later turn of the event loop.
    *
    *	process.nextTick is preferred, then setImmediate, then setTimeout
    *
    * @param func The function to be deferred.
    */
    TypedPromises.deferFunction = (function () {
        if (setImmediate)
            return setImmediate;

        return function (f) {
            setTimeout(f, 0);
        };
    })();

    /**
    *	This class represents a queue of functions that need to be iterated through.
    */
    var FunctionQueue = (function () {
        function FunctionQueue() {
            this._funcs = [];
            this._started = false;
        }
        /**
        *	Runs all functions currently in the FunctionQueue in the next turn of the event loop.
        */
        FunctionQueue.prototype._iterator = function () {
            var funcs = this._funcs;
            this._funcs = [];
            TypedPromises.deferFunction(function () {
                funcs.forEach(function (f) {
                    return f();
                });
            });
        };

        /**
        *	Enqueues a function in the FunctionQueue. If the iteration has been started, runs the iterator.
        *	@param func The function to defer.
        */
        FunctionQueue.prototype.enqueue = function (func) {
            if (isFunction(func)) {
                this._funcs.push(func);
                if (this._started && this._funcs.length === 1) {
                    this._iterator();
                }
            }
        };

        /**
        *	Start iteration through the functions.
        */
        FunctionQueue.prototype.start = function () {
            if (!this._started) {
                this._started = true;
                if (this._funcs.length > 0)
                    this._iterator();
            }
        };
        return FunctionQueue;
    })();

    /**
    * Creates a FunctionQueue.
    */
    function createFunctionQueue() {
        return new FunctionQueue();
    }

    /**
    * This class represents a deferral. It can be used to create a promise. Offers functions for fulfilling or rejecting a function.
    */
    var _deferral = (function () {
        function _deferral() {
            this._isPending = true;
            this._isResolved = false;
            this._isRejected = false;
            this._fulfillQueue = createFunctionQueue();
            this._rejectQueue = createFunctionQueue();
            var def = this;

            this.resolve = function (value) {
                return (isPromise(value)) ? value.then(function (v) {
                    return def._onResolve(v);
                }, function (r) {
                    return def._onReject(r);
                }) : def._onResolve(value);
            };

            this.reject = function (reason) {
                return def._onReject(reason);
            };

            this.promise = {
                _isAGPromise: true,
                then: function (onFulfilled, onRejected) {
                    var def2 = createDeferral();

                    def._queueFulfill(function () {
                        if (isFunction(onFulfilled)) {
                            try  {
                                var result = onFulfilled(def.promise._value);
                                def2.resolve(result);
                            } catch (e) {
                                def2.reject(e);
                            }
                        } else {
                            def2.resolve(def.promise._value);
                        }
                    });

                    def._queueReject(function () {
                        if (isFunction(onRejected)) {
                            try  {
                                var error = onRejected(def.promise._reason);
                                def2.resolve(error);
                            } catch (e) {
                                def2.reject(e);
                            }
                        } else {
                            def2.reject(def.promise._reason);
                        }
                    });

                    return def2.promise;
                }
            };
        }
        _deferral.prototype._onResolve = function (val) {
            if (this._isPending) {
                this._isResolved = true;
                this._isPending = false;

                this.promise._value = val;

                this._rejectQueue = null;
                this._fulfillQueue.start();
            }
        };

        _deferral.prototype._onReject = function (rea) {
            if (this._isPending) {
                this._isRejected = true;
                this._isPending = false;

                this.promise._reason = rea;

                this._fulfillQueue = null;
                this._rejectQueue.start();
            }
        };

        _deferral.prototype._queueFulfill = function (fulfillCallback) {
            if (this._fulfillQueue)
                this._fulfillQueue.enqueue(fulfillCallback);
        };

        _deferral.prototype._queueReject = function (rejectCallback) {
            if (this._rejectQueue)
                this._rejectQueue.enqueue(rejectCallback);
        };
        return _deferral;
    })();

    /**
    *	Creates a Deferral object.
    */
    function createDeferral() {
        return new _deferral();
    }
    TypedPromises.createDeferral = createDeferral;

    function promisify(valueOrPromise) {
        if (isAGPromise(valueOrPromise))
            return valueOrPromise;

        var def = createDeferral();
        def.resolve(valueOrPromise);
        return def.promise;
    }
    TypedPromises.promisify = promisify;

    /**
    Returns a TypedPromises promise rejected with the reason provided.
    
    @param reason The reason for the rejected promise
    */
    function rejectify(reason) {
        var def = createDeferral();
        def.reject(reason);
        return def.promise;
    }
    TypedPromises.rejectify = rejectify;
})(TypedPromises || (TypedPromises = {}));


module.exports = TypedPromises;

